///////////////////////////////////////////////////////////////////////////////
//
//    (C) Copyright 2009-2010 OSR Open Systems Resources, Inc.
//    All Rights Reserved
//
//    This work is distributed under the OSR Non-Commercial Software License
//    which is provided at
//    "http://www.osronline.com/page.cfm?name=NonCommLicense" in
//    the hope that it will be enlightening, but WITHOUT ANY
//    WARRANTY; without even the implied warranty of
//    MECHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//            OSR Open Systems Resources, Inc.
//            105 Route 101A Suite 19
//            Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//
//    MODULE:
//
//        uniqstack.cpp
//
//    ABSTRACT:
//
//       An implementation of !uniqstack for kernel mode
//
//    AUTHOR(S):
//
//        OSR Open Systems Resources, Inc.
// 
//    REVISION:   
//
//
///////////////////////////////////////////////////////////////////////////////
#include "apexts.h"

////////////// 
// TYPEDEFS //
////////////// 

// 
// This is the structure that we'll use to track unique thread
// stacks in the system
// 
#define MAX_FRAME_DEPTH 20

typedef struct _UNIQ_STACK {
    // 
    // A thread with the unique stack
    // 
    ULONG64           Thread;

    // 
    // The call stack
    // 
    DEBUG_STACK_FRAME StackFrames[MAX_FRAME_DEPTH];

    // 
    // Number of valid frames in StackFrames (< MAX_FRAME_DEPTH)
    // 
    ULONG             ValidFrames;
}UNIQ_STACK, *PUNIQ_STACK;



/////////////
// GLOBALS //
///////////// 
BOOLEAN     CountOnly;
ULONG       ThreadCount;
ULONG       UniqueThreads;
ULONG       ThreadsProcessed;
PUNIQ_STACK UniqStacks;



///////////////////////////////////////////////////////////////////////////////
//
//  uniqstack
//
//    Extension command that walks every thread in the system
//    and records any thread stacks that are unique. Uniqueness
//    is determined by grabbing up to MAX_FRAME_DEPTH stack
//    frames and recording the return instructions. A unique
//    sequence of return instructions is considered to be a
//    unique stack.
//
///////////////////////////////////////////////////////////////////////////////
HRESULT 
CALLBACK
uniqstack(
    PDEBUG_CLIENT4 Client, 
    PCSTR args
    ) {
    
    CHAR                  buffer[1024];
    PUNIQ_STACK           uniqStack;
    PDEBUG_CONTROL        debugControl;
    HRESULT               hr;
    ULONG                 threadCount;
    ULONG                 largestProcess;
    DEBUG_VALUE           startThread;


    UNREFERENCED_PARAMETER(args);

   
    // 
    // Extensions are single threaded, so we're OK with globals.
    // 
    ThreadCount      = 0;
    UniqueThreads    = 0;
    ThreadsProcessed = 0;

    // 
    // Our debug control instance will allow us to output, execute
    // commands, evaluate symbols, etc.
    // 
    hr = Client->QueryInterface(__uuidof(IDebugControl),
                                (VOID **)&debugControl);

    if (hr != S_OK) {
        // 
        // Can't do anything but silently fail at this point...
        // 
        return hr;
    }

    CommandBanner(debugControl);

    // 
    // We're going to leverage the !for_each_thread command to
    // enumerate the threads in the system. As the command, we'll
    // specify:
    // 
    // !threadcallback @#Thread
    // 
    // Which will call us back at our command threadcallback below
    // for each thread in the system with the address of the thread
    // specified as the parameter
    // 

    // 
    // First, we'll get the number of threads in the system
    // 
    CountOnly = TRUE;

    hr = debugControl->Execute(
                          DEBUG_OUTCTL_ALL_CLIENTS |
                          DEBUG_OUTCTL_OVERRIDE_MASK |
                          DEBUG_OUTCTL_NOT_LOGGED,
                          "!for_each_thread \"!threadcallback @#Thread\"",
                          DEBUG_EXECUTE_DEFAULT );
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't execute for each command (1) (0x%x)\n",
                             hr);
        debugControl->Release();
        return hr;
    }

    // 
    // No longer counting.
    // 
    CountOnly = FALSE;

    if (ThreadCount == 0) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Didn't count any threads\n");
        debugControl->Release();
        return S_OK;
    }

    debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                         DEBUG_OUTCTL_OVERRIDE_MASK |
                         DEBUG_OUTCTL_NOT_LOGGED,
                         "There are %d threads in the system\n",
                         ThreadCount);

    UniqStacks = (PUNIQ_STACK)malloc(ThreadCount * sizeof(UNIQ_STACK));
    if (UniqStacks == NULL) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't allocate memory for thread info\n");
        debugControl->Release();
        return S_OK;
    }


    // 
    // Our thread callback is going to change thread contexts with
    // .thread for each thread specified, so remember the thread we
    // started at so that we can switch back to it.
    // 
    hr = debugControl->Evaluate("@$thread",
                                DEBUG_VALUE_INT64,
                                &startThread,
                                NULL);
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't evaluate current thread (0x%x)\n",
                             hr);
        free(UniqStacks);
        debugControl->Release();
        return hr;
    }


    // 
    // Now call again to get the callstacks of all the unique
    // threads.
    // 
    hr = debugControl->Execute(
                          DEBUG_OUTCTL_ALL_CLIENTS |
                          DEBUG_OUTCTL_OVERRIDE_MASK |
                          DEBUG_OUTCTL_NOT_LOGGED,
                          "!for_each_thread \"!threadcallback @#Thread\"",
                          DEBUG_EXECUTE_DEFAULT);
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't execute for each command (2) (0x%x)\n",
                             hr);
        free(UniqStacks);
        debugControl->Release();
        return hr;
    }

    debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                         DEBUG_OUTCTL_OVERRIDE_MASK |
                         DEBUG_OUTCTL_NOT_LOGGED,
                         "But there are only %d unique threads in the system\n",
                         UniqueThreads);

    for (ULONG i = 0; i < UniqueThreads; i++) {

        uniqStack = &UniqStacks[i];

        hr = StringCbPrintf(buffer, 
                            sizeof(buffer),
                            "!thread 0x%I64x", 
                            uniqStack->Thread); 
        if (hr != S_OK) {
            debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                                 DEBUG_OUTCTL_OVERRIDE_MASK |
                                 DEBUG_OUTCTL_NOT_LOGGED,
                                 "Couldn't format command (0x%x)\n",
                                 hr);
            break;
        }
        hr = debugControl->Execute(DEBUG_OUTCTL_ALL_CLIENTS |
                                   DEBUG_OUTCTL_OVERRIDE_MASK |
                                   DEBUG_OUTCTL_NOT_LOGGED,
                                   buffer,
                                   DEBUG_EXECUTE_DEFAULT);
        if (hr != S_OK) {
            debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                                 DEBUG_OUTCTL_OVERRIDE_MASK |
                                 DEBUG_OUTCTL_NOT_LOGGED,
                                 "Couldn't execute !thread (0x%x)\n",
                                 hr);
            break;
        }

    }

    // 
    // Clean up.
    // 

    // 
    // Switch back to the original thread.
    // 
    hr = StringCbPrintf(buffer, 
                        sizeof(buffer),
                        ".thread 0x%I64x", 
                        startThread.I64); 
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't format command (0x%x)\n",
                             hr);
    }

    // 
    // Now execute the command to switch thread context.
    // 
    hr = debugControl->Execute(DEBUG_OUTCTL_IGNORE |
                               DEBUG_OUTCTL_OVERRIDE_MASK |
                               DEBUG_OUTCTL_NOT_LOGGED,
                               buffer,
                               DEBUG_EXECUTE_DEFAULT);
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't execute command (0x%x)\n",
                             hr);
    }

    free(UniqStacks);
    debugControl->Release();
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//
//  threadcallback
//
//    !uniqstack does the work of walking every thread in the
//    system by calling !for_each_thread. This command is given
//    as a callback to that command, thus it's called for each
//    thread in the system.
// 
//    This callback is called for every thread twice. First with
//    CountOnly set to TRUE to indicate that !uniqstack is just
//    getting a count of threads and then again to actually
//    capture the call stacks of the threads.
// 
///////////////////////////////////////////////////////////////////////////////
HRESULT 
CALLBACK
threadcallback(
    PDEBUG_CLIENT4 Client, 
    PCSTR args
    ) {

    CHAR              buffer[1024];
    DEBUG_STACK_FRAME stackFrames[MAX_FRAME_DEPTH];
    ULONG             validFrames;
    PUNIQ_STACK       uniqStack;
    ULONG             i;
    ULONG             j;
    HRESULT           hr;
    PDEBUG_CONTROL    debugControl;
    DEBUG_VALUE       threadAddr;

    if (CountOnly == TRUE) {
        // 
        // If we're just counting, then we'll just bump the count and
        // move on.
        // 
        ThreadCount++;
        return S_OK;
    }

    // 
    // Need a debug control here as well
    // 
    hr = Client->QueryInterface(__uuidof(IDebugControl),
                                (void **)&debugControl);

    if (hr != S_OK) {
        // 
        // Have to fail silently...
        // 
        return hr;
    }

    // 
    // The first argument to this command is a thread address. Let's
    // convert the string into a 64bit value
    // 
    hr = debugControl->Evaluate(args,
                                DEBUG_VALUE_INT64,
                                &threadAddr,
                                NULL);
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't convert argument to address (0x%x)\n",
                             hr);
        debugControl->Release();
        return hr;
    }

    // 
    // Now let's build a command to switch the thread context to
    // the given thread.
    // 
    hr = StringCbPrintf(buffer, 
                        sizeof(buffer),
                        ".thread 0x%I64x", 
                        threadAddr.I64); 
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't format command (0x%x)\n",
                             hr);
        debugControl->Release();
        return hr;
    }

    // 
    // Now execute the command to switch thread context.
    // 
    hr = debugControl->Execute(DEBUG_OUTCTL_IGNORE |
                               DEBUG_OUTCTL_OVERRIDE_MASK |
                               DEBUG_OUTCTL_NOT_LOGGED,
                               buffer,
                               DEBUG_EXECUTE_DEFAULT);
    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't execute command (0x%x)\n",
                             hr);
        debugControl->Release();
        return hr;
    }
    
    // 
    // Now we can get a stack trace for the current thread.
    // 
    hr = debugControl->GetStackTrace(0,
                                     0,
                                     0,
                                     &stackFrames[0],
                                     MAX_FRAME_DEPTH,
                                     &validFrames);

    if (hr != S_OK) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "Couldn't capture stack frames (0x%x)\n",
                             hr);
        debugControl->Release();
        return hr;
    }

    // 
    // And compare that to the threads that we've already found.
    // 
    for (i = 0; i < UniqueThreads; i++) {

        uniqStack = &UniqStacks[i];

        if (uniqStack->ValidFrames != validFrames) {
            //
            // No match.
            //
            continue;
        }

        for (j = 0; j < validFrames; j++) {
            // 
            // Just keep comparing the instructions on the stack and break
            // if we get a mismatch.
            // 
            if (uniqStack->StackFrames[j].InstructionOffset ==
                                    stackFrames[j].InstructionOffset) {
                continue;
            }
            break;
        }

        if (j == validFrames) {
            //
            // It's a match!
            //  
            break;
        }
    }

    if (i == UniqueThreads) {
        //
        // Didn't match this thread, add it.
        //

        // 
        // Grab a free unique stack structure
        // 
        uniqStack = &UniqStacks[UniqueThreads];

        // 
        // Copy the frames in.
        // 
        uniqStack->ValidFrames = validFrames;
        memcpy(&uniqStack->StackFrames[0],
               &stackFrames[0],
               (sizeof(DEBUG_STACK_FRAME) * validFrames));

        // 
        // Store away the thread address
        // 
        uniqStack->Thread = threadAddr.I64;

        // 
        // And increment the number of unique threads
        // 
        UniqueThreads++;

    }

    // 
    // Give some sort of progress indicator.
    // 
    ThreadsProcessed++;
    debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                         DEBUG_OUTCTL_OVERRIDE_MASK |
                         DEBUG_OUTCTL_NOT_LOGGED,
                         ".");
    if (ThreadsProcessed  == ThreadCount) {
        debugControl->Output(DEBUG_OUTCTL_ALL_CLIENTS |
                             DEBUG_OUTCTL_OVERRIDE_MASK |
                             DEBUG_OUTCTL_NOT_LOGGED,
                             "\n",
                             UniqueThreads);
    }

    debugControl->Release();
    return S_OK;
}


